/*
Algebra.h
Written by Walter O. Krawec and Eric Geiss
Copyright (c) 2017

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef _ALGEBRA_H_
#define _ALGEBRA_H_

#define USE_LINUX

#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <complex.h>

#include "Complex.hpp"

#define ComplexType double _Complex

namespace algebra
{

	class ComplexHelper
	{
	public:
	  static ComplexType one() {ComplexType out; out = 1.0 + 0*I; return out; }
	  static ComplexType zero() {ComplexType out; out = 0 + 0*I; return out;}

		static void sq(ComplexType& v)
		{
		  ComplexType v2 = creal(v) - cimag(v)*I;
		  v = mult(v, v2);
		}
		static ComplexType construct(double theta)
		{
		  ComplexType out = cos(theta) + sin(theta)*I;
		  return out;
		}

		static ComplexType power(ComplexType& base, int e)
		{
		  return cpow(base, e);
		}

		static ComplexType add(ComplexType& a, ComplexType& b)
		{
		  return a+b;
		}
		static ComplexType subtract(ComplexType& a, ComplexType& b)
		{
		  return a-b;
		}

		static ComplexType mult(ComplexType& a, ComplexType& b)
		{
		  return a*b;
		}

		static ComplexType invert(ComplexType& a)
		{
			ComplexType out;
			double d = creal(a)*creal(a) + cimag(a)*cimag(a);

			out = creal(a) / d - cimag(a)/d *I;

			return out;
		}
	};

	class matrix
	{
	public:
		matrix()
		{
			elements = NULL;
			noCleanup = false;
			r = 0;
			c = 0;
			n = 0;
		}
		matrix(const char* _name)
		{
			elements = NULL;
			noCleanup = false;
			r = 0;
			c = 0;
			n = 0;
			varName = std::string(_name);
		}
		matrix(const matrix& m)
		{
			elements = NULL;
			r = m.r;
			c = m.c;
			n = r*c;

			if(r == 0 || c == 0 || !m.elements)
				return;

			elements = new ComplexType[n];

			noCleanup = false;

			for(int i=0; i<n; ++i)
				elements[i] = m.elements[i];
		}
		~matrix() {cleanup();}

		void saveToFile(const std::string& filename);
		void saveToFile(std::ofstream& f);

		bool loadFromFile(const std::string& filename);
		bool loadFromFile(std::ifstream& f);


		matrix& operator=(const matrix& m)
		{
			if(!elements || r != m.r || c != m.c)
			{
				cleanup();
				r = m.r;
				c = m.c;
				n = r*c;

				if(r == 0 || c == 0 || !m.elements)
				{
				  std::cout << "operator= error.\n";
					throw std::string("operator= error.\n");
					return *this;
				}

				elements = new ComplexType[n];
			}

			if(!m.elements)
			{
			  std::cout << "operator= error.\n";
				throw std::string("operator= error 2.\n");
				return *this;
			}

			for(int i=0; i<n; ++i)
				elements[i] = m.elements[i];

			noCleanup = false;

			return *this;
		}

		void cleanup()
		{
			if(elements && !noCleanup)
			{
				delete[] elements;
			}
			elements = NULL;
			r = 0;
			c = 0;
			n = 0;
		}

		void zero()
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
			  elements[i] = 0.0+0.0*I;
			}
		}

		void ident()
		{
			if(!elements)
				return;

			zero();
			int dim = getRows();
			if(getCols() < getRows())
				dim = getCols();

			ComplexType one = 1.0+0.0*I;

			for(int i=0; i<dim; ++i)
				this->setDirect(i,i,one);
		}

		bool eigenValues(std::vector <double>& values);

		ComplexType trace()
		{
			ComplexType c = 0.0 + 0.0*I;
			for(int i=0; i<getRows(); ++i)
			{
			  c = c + (*this)(i,i);
			}
			return c;
		}

		double traceRe()
		{
		  ComplexType tr = trace();
		  return creal(tr);
		}

		void clear(ComplexType value)
		{
			if(!elements)
				return;
			for(int i=0; i<n; ++i)
				elements[i] = value;
		}

		bool create(int _r, int _c, ComplexType* _elements)
		{
			cleanup();
			r = _r;
			c = _c;
			n = r*c;

			elements = _elements;
			noCleanup=true;
			return true;
		}
		bool create(int _r, int _c)
		{
			if(_r <= 0 || _c <= 0)
				return false;
			if(elements && r == _r && c == _c)
				return true;

			cleanup();
			r = _r;
			c = _c;
			n = r*c;

			elements = new ComplexType[r*c];

			noCleanup = false;

			return true;
		}

		bool create(int _r)
		{
			return create(_r, _r);
		}

		ComplexType& operator() (int row, int col)
		{
			int _r = row;
			int _c = col;
			if(!elements || _r >= r || _c >= c || _r < 0 || _c < 0)
			{
				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				std::cout << "Name = " << varName << "\n";
				throw(std::string("Algebra Error 5.1"));
				//throw(5);
				//return double();
			}

			return elements[_c*r + _r];
		}

		ComplexType getDirect(int row, int col)
		{
			int _r = row;
			int _c = col;
			if(!elements || _r >= r || _c >= c || _r < 0 || _c < 0)
			{
				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				throw(std::string("Algebra Error 5.2"));
				//throw(5);
				//return double();
			}

			return elements[_c*r + _r];
		}

		void setDirect(int _n, ComplexType data)
		{
			if(!elements || _n >= n)
			{
//				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				throw(std::string("Algebra Error 5.3"));
				//throw(5);
				//return double();
			}

			elements[_n] = data;
		}

		void setDirect(int row, int col, ComplexType data)
		{
			int _r = row;
			int _c = col;
			if(!elements || _r >= r || _c >= c || _r < 0 || _c < 0)
			{
				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				throw(std::string("Algebra Error 5.4"));
				//throw(5);
				//return double();
			}

			elements[_c*r + _r] = data;
		}

		void setDirect(int row, int col, double re, double im)
		{
			int _r = row;
			int _c = col;
			if(!elements || _r >= r || _c >= c || _r < 0 || _c < 0)
			{
				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				throw(std::string("Algebra Error 5.4"));
				//throw(5);
				//return double();
			}

			elements[_c*r + _r] = re + im*I;
		}
	        ComplexType get(int _n)
		{
			if(!elements || _n >= n)
			{
				std::cout << "Error 2 (" << elements << ", " << _n << ", " << n << ")\n";
				return 0;
			}

			return elements[_n];
		}

		ComplexType* getElements()
		{
			return elements;
		}

		void print(std::ostream& f, bool printAll=false)
		{
			if(!elements)
				return;
			int rStart = 0, cStart=0;
		//	if(r > 10)
			//	rStart = r-10;
		//	if(c > 10)
		//		cStart = c-10;

			int rows = alg_min(10, r);
			int cols = alg_min(10, c);

			if(printAll)
			{
				rows = r;
				cols = c;
			}
			for(int _r=rStart; _r < rows; ++_r)
			{
				for(int _c=cStart; _c<cols; ++_c)
				{
				  f << creal((*this)(_r,_c)) << "+" << cimag((*this)(_r,_c)) << "i\t";
				}
				f << "\n";
			}

			f << "\n";
		}

		void multiplyScalar(ComplexType& c)
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
				elements[i] = ComplexHelper::mult(elements[i], c);
			}
		}

		void multiplyScalar(double c)
		{
		  if(!elements)
		    return;
		  for(int i=0; i<n; ++i)
		    elements[i] *= c;
		}

		void sqTranspose()
		{
			if(!elements || c != r)
			{
				throw(std::string("Algebra error 1.2"));
				return;
			}

			ComplexType temp;
			for(long row=0; row<r; ++row)
			{
				for(long col=row; col<c; ++col)
				{
					if(row == col)
						continue;
					temp = getDirect(row,col);
					setDirect(row,col, getDirect(col,row));
					setDirect(col,row, temp);
				}
			}
		}

		void transpose(matrix& B)
		{
			if(!elements)
				return;
			B.create(c,r);
			for(int _r=0; _r<r; ++_r)
			{
				for(int _c=0; _c<c; ++_c)
				{
				  ComplexType c = (*this)(_r, _c);
				  B(_c, _r) = creal(c) - cimag(c)*I;
				}
					//B.set(_c,_r, get(_r,_c));
			}
		}

		void subtract(ComplexType& c)
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
			  elements[i] = elements[i] = c;
			}
		}

		void subtract(matrix* B, matrix* output)
		{
			if(!B || !output)
				return;

			if(!elements || !B->getElements() || getRows() != B->getRows() || getCols() != B->getCols())
			{
				throw std::string("Error - cannot subtract matrix\n");
				return;
			}

			output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
			  output->elements[i] = elements[i] - B->getElements()[i];
			}
		}

		// set ouput = sum_i pi*Mi
		static void addMulti(matrix& M1, double p1, matrix& M2, double p2, matrix& M3, double p3, matrix& M4, double p4, matrix& output)
		{
		  // todo: check sizes correct
		  output.create(M1.getRows(), M1.getCols());
		  for(int i=0; i<M1.n; ++i)
		    output.elements[i] = p1*M1.elements[i] + p2*M2.elements[i]
		      + p3*M3.elements[i] + p4*M4.elements[i];
		}


		void add(matrix* B, matrix* output)
		{
			if(!B || !output)
				return;

			if(!elements || !B->getElements() || getRows() != B->getRows() || getCols() != B->getCols())
			{
				throw std::string("Error - cannot add matrix\n");
				return;
			}

			output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
			  output->elements[i] = elements[i] + B->getElements()[i];
			}
		}

		void random(double smallValue, double largeValue);
		void add(matrix* B)
		{
			matrix* output = this;
			if(!B)
				return;

			if(!elements || !B->getElements() || getRows() != B->getRows() || getCols() != B->getCols())
			{
				throw std::string("Error - cannot add matrix\n");
				return;
			}

			//output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
			  output->elements[i] = elements[i] + B->getElements()[i];
			}
		}

		// adds: this = this + c2*B
		void add(quantum::Complex& c2, matrix& B)
		{
#ifdef USE_LINUX
		  ComplexType c = c2.getReal() + I*c2.getImaginary();
#else

#endif
			matrix* output = this;

			if(!elements || !B.getElements() || getRows() != B.getRows() || getCols() != B.getCols())
			{
				throw std::string("Error - cannot add matrix\n");
				return;
			}

			//output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
			  output->elements[i] = elements[i] + ComplexHelper::mult(c, B.getElements()[i]);
			}
		}


		void dot(matrix* B, ComplexType& output);	// returns this <dot> B
		// set Trans* = 'T' for *^T
		bool multiply(matrix& B, matrix& output, char TRANSB='N');

		void tensor(matrix* B, matrix* output);

		int getRows() const
		{
			return r;
		}
		int getCols() const
		{
			return c;
		}
		int getSize()
		{
			return n;
		}

		double entropy()
		{
		  std::vector<double> ev;
		  double sum = 0;
		  
		  eigenValues(ev);
		  
		  for(int i=0; i < ev.size(); ++i)
		    {
		      sum = sum - ev[i] * safeLog(ev[i]);
		    }

		  return sum;
		}
	private:
		static double safeLog(double x)
		{
		  if(x < 0.0000001)
		    return 0;
		  else
		    return log(x)/log(2.0);
		}

		int alg_min(int a, int b)
		{
			return (a < b) ? a : b;
		}
		int alg_max(int a, int b)
		{
			return (a > b) ? a : b;
		}
		ComplexType* elements;
		int r, c, n;

		bool noCleanup;

		std::string varName;
	};

	typedef matrix mat;
};

#endif
